'''
Manages a set of instantaneous data to plot the current sentiment on a gauge (with variable weight).
'''

import datetime
import functools
from enum import Enum

import ipywidgets as widgets
import numpy as np
from finndex.aggregate import sliders
from finndex.graphing import gauge
from finndex.sentiment import fearandgreed, nlp, trends
from finndex.util import dateutil, mathutil, webutil
from IPython.display import display

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"


# Represents a static named data reading represented by a single gauge and weighted by a slider.
class DataReading:
   def __init__(self, name, gauge, slider):
      self.name = name
      self.gauge = gauge
      self.slider = slider

'''
Represents all possible values that can be plotted instantaneously. Each value corresponds to a standard data retrieval function.
'''
class StaticMetricType(Enum):
   FEAR_AND_GREED = functools.partial(fearandgreed.displayFearAndGreedDate)
   TRENDS = functools.partial(trends.displayTrendsDate)

'''
Computes and displays on a gauge instantaneous sentiment value given a set of keywords. Weights can be modified using sliders;
if weights are provided in the 'weights' parameter, presents a static gauge using those weights. If the 'displayIntermediates'
parameter is True, displays gauges for each individual value above the aggregate gauge.
'''
class InstantaneousSentimentManager:
   def __init__(self, keywordsList, date=dateutil.getCurrentDateTime(), weights=None, displayIntermediates=False):
      self.dataVals = []
      slidersList = []

      for idx, keyword in enumerate(keywordsList):
         dataReading = DataReading(name=str(keyword), gauge=keyword.value(date=date, display=displayIntermediates), 
                                   slider=sliders.Slider(str(keyword), widgets.FloatSlider(min=0.0, max=sliders.MAX_VAL, 
                                                                                           step=sliders.STEP, 
                                                                                           value=weights[idx] if weights != None else 0.0)))
         self.dataVals += [dataReading]
         slidersList += [dataReading.slider]



      self.gauge = None


      if weights == None:
         self.sliderManager = sliders.SliderManager(self.updateGauge, slidersList)
      else:
         self.sliderManager = None

   '''
   Retrieves the data reading represented by a given name.
   '''
   def getDataReading(self, name):
      return [reading for reading in self.dataVals if reading.name == name][0]

   '''
   Computes the aggregate sentiment based on a dictionary of values where a given value corresponds to a dictionary containing 
   the value's weight (with key 'weight'), the minimum possible value the value can take on (with key 'minVal'), and the 
   maximum possible value the value can take on (with key 'maxVal').

   Returns a tuple containing (in the following order) the aggregate value, the minimum value for this aggregate, 
   and the maximum value for this aggregate.
   '''
   def computeAggregateSentiment(self):
      aggregateVal = 0
      minVal = 0
      maxVal = 1

      for data in self.dataVals:
         weight = data.slider.getReading()
         scaledValue = mathutil.map(data.gauge.currentVal, data.gauge.minVal, data.gauge.maxVal, 0, 1)
         aggregateVal += scaledValue * weight
         
      return (aggregateVal, minVal, maxVal)

   '''
   Displays the aggregate sentiment meter based on a dictionary of values. See computeAggregateSentiment for a complete
   description of this parameter.
   '''
   def displayGauge(self):
      aggregateVal, minVal, maxVal = self.computeAggregateSentiment()
      
      self.gauge = gauge.Gauge(labels=['—','0','+'], 
                  colors=['#c80000', '#646400', '#00c800'], currentVal=aggregateVal, 
                  minVal = minVal,
                  maxVal = maxVal, title='Aggregate Sentiment')

      if self.sliderManager != None:
         display(self.sliderManager.generateDisplayBox())

      return self.gauge

   # Updates the corresponding gauge given a new set of values and the existing gauge.
   def updateGauge(self):
      self.gauge.currentVal, self.gauge.minVal, self.gauge.maxVal = self.computeAggregateSentiment()
      self.gauge.generateGauge()

   # Retrieves the aggregate sentiment value as a float.
   def getReading(self):
      return self.computeAggregateSentiment()



'''
Uses the CoinMetrics API to retrieve several useful pieces of fundamental data on cryptocurrencies.
'''

import datetime
import json
from enum import Enum

from finndex.graphing import timeseries
from finndex.util import cryptocurrencies, dateutil, mathutil, webutil

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

COIN_METRICS_API_PREFIX = "https://community-api.coinmetrics.io/v2/"
NETWORK_METRIC_SUFFIX = "assets/{}/metricdata?metrics="

COIN_METRICS_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

# Represents an enum containing several possible keywords which can be used with the CoinMetrics API.
class CoinMetricsData(Enum):
    BLOCK_COUNT = "BlkCnt"
    MARKET_CAP = "CapRealUSD"
    PRICE_USD = "PriceUSD"
    TRANSACTION_CNT = "TxCnt"
    DAILY_ADDRESSES = "AdrActCnt"
    
'''
Retrieves a dictionary containing the CoinMetrics data across all time for a given set of statistics for a given
set of cryptocurrencies. 

Valid and retrievable statistics are listed in the CoinMetricsData class.

Returns a dictionary containing a key 'metrics' pointing to all statistics which were retrieved and a key 'series' 
which points to an list of dictionaries. Each dictionary in this list contains a timestamp (key 'time') formatted to
the microsecond and a key 'values' pointing to a list of string values. The list is ordered in the same fashion
as the 'metrics' list.
'''
def getCoinMetricsDict(currenciesList, metricsList):
    returnDict = {}

    for currency in currenciesList:
        returnDict[currency] = {}
        
        desiredMetrics = COIN_METRICS_API_PREFIX + NETWORK_METRIC_SUFFIX.format(currency.value.lower())
        for keyword in metricsList:
            desiredMetrics += keyword.value + ","
        desiredMetrics = desiredMetrics[:-1] # remove final comma
        
        loadedPageData = json.loads(webutil.getPageContent(desiredMetrics))['metricData']
        metricsListRetrieved = loadedPageData['metrics']
        dataSet = loadedPageData['series']
        
        for dataDict in dataSet:
            valueDict = {CoinMetricsData(metricsListRetrieved[idx]):float(value) 
                     for idx, value in enumerate(dataDict['values'])}
            returnDict[currency][datetime.datetime.strptime(dataDict['time'], 
                                                                  COIN_METRICS_TIMESTAMP_FORMAT)] = valueDict
    return returnDict

'''
Retrieves from CoinMetrics a set of metrics (from [0, 1]) of a given set of currencies (type Cryptocurrencies) 
from a given date range.
'''
def getCoinMetricsDateRange(metric, startDate, endDate, currenciesList):
    dataDict = getCoinMetricsDict(currenciesList, [metric])
    
    newDict = {}
    for currency, values in dataDict.items():
        numMetrics = len(values[list(values.keys())[0]])
        
        minVal = min(val[metric] for val in values.values()) 
        maxVal = max(val[metric] for val in values.values())
            
        newDict[currency] = {}
        
        for date, metricsDict in values.items():
            if date.date() >= startDate.date() and date.date() <= endDate.date():
                newDict[currency][date] = mathutil.map(metricsDict[metric], minVal, maxVal, 0, 1) 
                
    return newDict
    
# Plots all data available from CoinMetrics across time for a given currency.
def plotAllCoinMetricsData(currency, startDate=dateutil.getCurrentDateTime() - datetime.timedelta(days=1000), endDate=dateutil.getCurrentDateTime()):
    for dataKey in list(CoinMetricsData):
        data = getCoinMetricsDateRange(dataKey, startDate, endDate, [currency])[currency.value]
        timeseries.TimeSeries("{}: {}".format(str(currency), str(dataKey)), {dataKey: {date:values[list(values.keys())[0]] for date, values in data.items()}})



'''
Sets up the library by starting the NLP sever and initializing the graphing engine's date format.
'''

from finndex.api import datadict
from finndex.sentiment import nlp
from pandas.plotting import register_matplotlib_converters

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"


nlp.startServer(nlp.STANFORD_NLP_TIMEOUT, nlp.STANFORD_NLP_PORT)
register_matplotlib_converters() # register MatPlotLib date converter

a = datadict.setupApi('__main__')



'''
Contains utility functions providing an extension to Python's datetime.
'''

import datetime

import pytz

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

DESIRED_DATE_FORMAT = "%Y-%m-%d"
DESIRED_TIME_ZONE = pytz.timezone("US/Pacific")

'''
Converts a timestamp (represented as a string) in one date format into another date format. Returns
the newly formatted date as a string.

A list of acceptable date format characters can be found at the following link.
https://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
'''
def convertTimestamp(timestamp, initialFormat, desiredFormat):
    return datetime.datetime.strptime(timestamp, initialFormat).strftime(desiredFormat)

'''
A generator function (inclusive on both endpoints) to iterate through every date or datetime in a given range.
'''
def dateRange(startDate, endDate):
   for i in range(int((endDate - startDate).days) + 1):
      yield startDate + datetime.timedelta(i)

'''
Retrieves the current time and date in Pacific time.
'''
def getCurrentDateTime():
   return datetime.datetime.now(DESIRED_TIME_ZONE)



'''
Contains a TimeSeries class allowing for a set of historical data to be plotted over time.
'''

import datetime

import matplotlib
import numpy as np
from finndex.util import dateutil
from matplotlib import cm
from matplotlib import pyplot as plt

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"


'''
Represents an easily modifiable time series. 

The data is provided in 'data', where each key (string) represents the type of value stored (like 'price' or 'sentiment')
and the corresponding value is an list of dictionaries where the key is date and the value is the corresponding value on that day.
Each list represents a single axis, thus all the data in a list will be plotted on the same axis. Up to 2 axes are supported.

dataDateFormat (string) represents the format of all the incoming x-values. graphDateFormat (string) represents the format in which the given
dates will be displayed on the x-axis. 
'''
class TimeSeries:
    def __init__(self, title, data, colors=['tab:red', 'tab:blue', 'tab:green','tab:purple', 'tab:pink', 'tab:orange'], dataDateFormat=dateutil.DESIRED_DATE_FORMAT, graphedDateFormat = None, yMin=None, yMax=None, fileName=False):
        self.title = title
        self.data = data
        self.dataDateFormat = dataDateFormat

        if graphedDateFormat == None:
            firstDataSet = data[list(data.keys())[0]]
            timeDifference = max(firstDataSet[list(firstDataSet.keys())[0]]) - min(firstDataSet[list(firstDataSet.keys())[0]])

            if timeDifference <= datetime.timedelta(days=30*6):
                graphedDateFormat = "%m-%d"
            elif timeDifference <= datetime.timedelta(days=30*12):
                graphedDateFormat = "%Y-%m"
            else:
                graphedDateFormat = "%Y"

        self.graphedDateFormat = graphedDateFormat
        self.yMin = yMin
        self.yMax = yMax
        self.colors=colors

        self.fig = None
        self.axes = []
        
        self.fileName = fileName 

        self.plotTimeSeries()

    
    '''
    Generates the graph using the data specified in the constructor, or, if already generated, updates it to a new data set.
    '''
    def plotTimeSeries(self):
        colors = list(self.colors) # duplicate so that colors can be removed as they are used
        if self.fig == None: # generating the graph for the first time
            self.fig, baseAxis = plt.subplots()
            firstExecution = True
        else:
            baseAxis = self.axes[0]
            firstExecution = False
        
        for idx, (valueType, valDictList) in enumerate(self.data.items()):
            if not firstExecution:
                desiredAxes = self.axes[idx]
            else:
                if idx == 0:
                    desiredAxes = baseAxis
                else:
                    desiredAxes = baseAxis.twinx()
                self.axes += [desiredAxes]
                
            
            desiredAxes.clear()

            desiredAxes.set_ylabel(valueType, color=self.colors[idx])

            for key, valDict in valDictList.items():
                dates = [date for date in valDict]
                values = [val for val in valDict.values()]

                formattedDates = []
                for date in dates:
                    if isinstance(date, str):
                        formattedDates += [datetime.datetime.strptime(date, self.dataDateFormat)]
                    elif isinstance(date, datetime.date):
                        formattedDates += [datetime.datetime.combine(date, datetime.time())]
                    else:
                        formattedDates += [date]
                    
                dates = matplotlib.dates.date2num(formattedDates)
                desiredAxes.plot(formattedDates, values, color = colors[0], label=str(key))

                del colors[0] # remove first color to shift all left
                
            if len(valDictList) > 1: # multiple curves plotted on axes
                desiredAxes.legend()

            desiredAxes.set_title(self.title)
            desiredAxes.xaxis.set_major_formatter(matplotlib.dates.DateFormatter(self.graphedDateFormat))
            
            if self.yMin != None:
                desiredAxes.set_ylim(ymin=self.yMin)
            if self.yMax != None:
                desiredAxes.set_ylim(ymax=self.yMax)
            
        self.fig.tight_layout()
        plt.show()

        if self.fileName:
            self.fig.savefig(self.fileName, dpi=200)



'''
Uses MatPlotLib to display a gauge which plots a piece of data with a set minimum and maximum.
'''

import os
import sys

import matplotlib
import numpy as np
from finndex.util import mathutil
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.patches import Circle, Rectangle, Wedge

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

ARROW_TAIL_RADIUS = 0.015

# Determines the range of angles which represent each portion of a gauge with n items.
def degreeRange(n): 
    start = np.linspace(0,180,n+1, endpoint=True)[0:-1]
    end = np.linspace(0,180,n+1, endpoint=True)[1::]
    mid_points = (start + end)/2.0
    return np.c_[start, end], mid_points

# Rotates a piece of text by a given angle.
def rot_text(ang): 
    rotation = np.degrees(np.radians(ang) * np.pi / np.pi - np.radians(90))
    return rotation

'''
Represents a gauge with a set of labels and colors (formatted in hex) corresponding to each of those labels.
'''
class Gauge:
    def __init__(self, labels, colors, currentVal, minVal, maxVal, title='', displayGauge=True, fileName=False):
        self.labels = labels
        self.colors = colors
        self.currentVal = currentVal
        self.minVal = minVal
        self.maxVal = maxVal
        self.title = title
        
        self.gaugeGenerated = False

        self.fig = None
        self.ax = None

        self.fileName = fileName
        
        if displayGauge:
            self.generateGauge()
        
        
    '''
    Creates and displays a gauge. Returns a tuple containing the generated figure, axes,
    and arrow.
    '''
    def generateGauge(self):
        if not self.gaugeGenerated:
            N = len(self.labels)

            if isinstance(self.colors, list): 
                self.colors = self.colors[::-1]

            """
            begins the plotting
            """

            self.fig, self.ax = plt.subplots()

            ang_range, mid_points = degreeRange(N)

            self.labels = self.labels[::-1]

            """
            plots the sectors and the arcs
            """
            patches = []
            for ang, c in zip(ang_range, self.colors): 
                # sectors
                patches.append(Wedge((0.,0.), .4, *ang, facecolor='w', lw=2))
                # arcs
                patches.append(Wedge((0.,0.), .4, *ang, width=0.10, facecolor=c, lw=2, alpha=0.5))

            [self.ax.add_patch(p) for p in patches]


            """
            set the labels (e.g. 'LOW','MEDIUM',...)
            """

            for mid, lab in zip(mid_points, self.labels):
                AVG_STR_LENGTH = 12
                AVG_FONT_SIZE = 11
                MAX_FONT_SIZE = 30

                fontSize = min(AVG_STR_LENGTH / len(lab) * AVG_FONT_SIZE, MAX_FONT_SIZE) # create dynamic font size based on string length
                self.ax.text(0.35 * np.cos(np.radians(mid)), 0.35 * np.sin(np.radians(mid)), lab,
                    horizontalalignment='center', verticalalignment='center', fontsize=fontSize,
                    fontweight='bold', rotation = rot_text(mid))

            """
            set the bottom banner and the title
            """
            r = Rectangle((-0.4,-0.1),0.8,0.1, facecolor='w', lw=2)
            self.ax.add_patch(r)

            self.ax.text(0, -0.05, self.title, horizontalalignment='center',
                verticalalignment='center', fontsize=22, fontweight='bold')

            #Plot the arrow based on given sentiment value.
            lowestAngle = ang_range[0][0]
            highestAngle = ang_range[-1][1]

            pos = mathutil.map(self.currentVal, self.minVal, self.maxVal, highestAngle, lowestAngle)

            self.arrow = self.ax.arrow(0, 0, 0.225 * np.cos(np.radians(pos)), 0.225 * np.sin(np.radians(pos)),
                        width=0.03, head_width=0.09, head_length=0.1, fc='k', ec='k')

            self.ax.add_patch(Circle((0, 0), radius=ARROW_TAIL_RADIUS, facecolor='k')) # make the arrow rounded at the tail

            """
            removes frame and ticks, and makes axis equal and tight
            """

            self.ax.set_frame_on(False)
            self.ax.axes.set_xticks([])
            self.ax.axes.set_yticks([])
            self.ax.axis('equal')

            plt.tight_layout()
            plt.show()
        else:
            self.arrow.remove()
            highestAngle = 180
            lowestAngle = 0
            pos = mathutil.map(self.currentVal, self.minVal, self.maxVal, highestAngle, lowestAngle)
            self.arrow = self.ax.arrow(0, 0, 0.225 * np.cos(np.radians(pos)), 0.225 * np.sin(np.radians(pos)),
                                                           width=0.03, head_width=0.09, head_length=0.1, fc='k', ec='k')
      
            self.fig.canvas.draw_idle()
        
        self.gaugeGenerated = True

        if self.fileName:
            self.fig.savefig(self.fileName, dpi=200)

        return (self.fig, self.ax, self.arrow)

# Generates a gauge with options "Low", "Medium", and "High".
def displayNeutralGauge(currentVal, minVal, maxVal, title, display=True):
    return Gauge(labels=["Low", "Medium", "High"], colors=['#c80000','#646400','#00c800'], currentVal=currentVal,
                 minVal=minVal, maxVal=maxVal, title=title, displayGauge=display)



'''
Provides utility functions to assist with number manipulation.
'''

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

# Linearly maps a value within one range to a new range.
def map(num, init_min, init_max, new_min, new_max):
    return (num - init_min) * (new_max - new_min) / (init_max - init_min) + new_min



'''
Extracts data from the Fear and Greed API and displays it instantaneously and provides a historical representation in a dictionary.
'''

import datetime
import json

from finndex.graphing import gauge
from finndex.util import cryptocurrencies, dateutil, mathutil, webutil

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

FEAR_AND_GREED_ADDRESS = "https://api.alternative.me/fng/?limit=0&date_format=us"
    
MIN_FEAR_AND_GREED = 0
MAX_FEAR_AND_GREED = 100

# Uses the Fear and Greed API to extract the Fear and Greed value from any given date.
def getFearAndGreed(date):
    timestampFormatted = date.strftime(dateutil.DESIRED_DATE_FORMAT) 
    
    return getAllFearAndGreed()[timestampFormatted]
   
'''
Uses the Fear and Greed API to extract all Fear and Greed values available as a range from 0-1. Returns a dictionary with key as date
and value the Fear and Greed value on that date.
'''
def getAllFearAndGreed():
    fearAndGreedVals = webutil.getPageContent(FEAR_AND_GREED_ADDRESS)
    jsonUnpacked = json.loads(fearAndGreedVals)
    dataArr = jsonUnpacked['data']
    dataDict = {}
    for singleDay in dataArr:
        timestampFormatted = dateutil.convertTimestamp(singleDay['timestamp'], '%m-%d-%Y', dateutil.DESIRED_DATE_FORMAT)
        dataDict[timestampFormatted] = mathutil.map(int(singleDay['value']), MIN_FEAR_AND_GREED, MAX_FEAR_AND_GREED, 0, 1)
        
    return dataDict

'''
Uses the Fear and Greed API to extract all Fear and Greed values available in a given date range. Returns a dictionary with key as date
and value the Fear and Greed value on that date.
'''
def getFearAndGreedDateRange(startDate, endDate, currenciesList=[cryptocurrencies.Cryptocurrencies.BITCOIN]):
    fearAndGreedVals = getAllFearAndGreed()
    dataDict = {}

    for date, data in fearAndGreedVals.items():
        timestampDate = datetime.datetime.strptime(date, '%Y-%m-%d')

        if timestampDate.date() >= startDate.date() and timestampDate.date() <= endDate.date(): # only consider year, month, and day
            dataDict[timestampDate.date()] = data
        
    return dataDict

# Displays a given Fear and Greed value (0-100) in a convenient gauge format.
def displayFearAndGreedNum(val, display=True):
    return gauge.Gauge(labels=['Extreme Fear','Fear','Greed','Extreme Greed'], 
      colors=['#c80000','#c84b00','#64a000', '#00c800'], currentVal=val, minVal = 0,
                 maxVal = 1, title='Fear and Greed', displayGauge=display)

# Displays the Fear and Greed value from a given date in a convenient gauge format.
def displayFearAndGreedDate(date=dateutil.getCurrentDateTime(), display=True):
    return displayFearAndGreedNum(getFearAndGreed(date), display)



'''
Uses Stanford's NLP library to analyze and the sentiment of a block of text and provides functions to stop and start the
server.
'''

import os

import numpy
from pycorenlp import StanfordCoreNLP

from finndex.graphing import gauge

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

STANFORD_NLP_LOCATION = '~/stanford-corenlp-full-2018-10-05/' # the location of the Stanford NLP library on my computer 
                                                              # (download at https://stanfordnlp.github.io/CoreNLP)
STANFORD_NLP_TIMEOUT = 100000 # the time after which a given NLP request will be killed if not yet complete
STANFORD_NLP_PORT = 9002

MIN_SENTIMENT = 0
MAX_SENTIMENT = 4

# Starts the Stanford NLP server with a given timeout and port.
def startServer(timeout, port):
   os.popen('cd {}; java -mx5g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -timeout {} -port {} &>/dev/null'.format(
         STANFORD_NLP_LOCATION, timeout, port)) 

# Determines the sentiment of a given block of text as a range from 0 (Extremely Negative) to 4 (Extremely Positive).
def findSentiment(text):
    NLP_SERVER_LOCATION = 'http://localhost:{}'.format(STANFORD_NLP_PORT)
    PROPERTIES_DICTIONARY = {'annotators': 'sentiment', 'outputFormat': 'json', 'timeout': STANFORD_NLP_TIMEOUT}
    
    nlp = StanfordCoreNLP(NLP_SERVER_LOCATION)
    result = nlp.annotate(text, properties = PROPERTIES_DICTIONARY)

    sentiments = []
    for sentenceAnalysis in result['sentences']:
        sentimentValue = float(sentenceAnalysis['sentimentValue'])
        sentiments += [sentimentValue]
    return numpy.average(sentiments)

# Displays a sentiment value (0-4) in a convenient gauge format.
def displaySentimentNum(sentimentVal):
    return gauge.Gauge(labels=['Very Negative','Negative','Neutral','Positive', 'Very Positive'], 
      colors=['#c80000','#c84b00','#646400','#64a000', '#00c800'], currentVal=sentimentVal, minVal = MIN_SENTIMENT, maxVal = MAX_SENTIMENT, title='Cryptocurrency Sentiment')

# Computes the sentiment value of a given piece of text and displays it as a gauge.
def displaySentimentTxt(text):
    return displaySentimentNum(findSentiment(text))    

# Kills the server running on a given port. 
def stopServer(port):
   exec('kill $(lsof -ti tcp:{})'.format(port))



import datetime

from finndex.aggregate import historical, instantaneous
from finndex.util import dateutil

import flask
import flask_restful

HISTORICAL_SENTIMENT_KEY = 'histSentiment'
INSTANTANEOUS_SENTIMENT_KEY = 'instSentiment'

API_OK_VAL = 200
API_FAILED_VAL = 404

class DataInfo(flask_restful.Resource):
   def get(self, data):
      if data == HISTORICAL_SENTIMENT_KEY:
         historicalManager = historical.HistoricalSentimentManager([historical.HistoricalMetricType.FEAR_AND_GREED,
                                                           historical.HistoricalMetricType.TRENDS,
                                                           historical.HistoricalMetricType.BLOCK_COUNT,
                                                           historical.HistoricalMetricType.DAILY_ADDRESSES,
                                                           historical.HistoricalMetricType.TRANSACTION_CNT],
                                                          startDate = dateutil.getCurrentDateTime() - datetime.timedelta(days=60),
                                                          endDate = dateutil.getCurrentDateTime(),
                                                         unweightedKeywordsList=[historical.HistoricalMetricType.PRICE_USD])
         return historicalManager.getHistoricalSentiment(), API_OK_VAL

      elif data == INSTANTANEOUS_SENTIMENT_KEY:
         instantaneousManager = instantaneous.InstantaneousSentimentManager([instantaneous.StaticMetricType.FEAR_AND_GREED, 
                                                       instantaneous.StaticMetricType.TRENDS])
         return instantaneousManager.getReading(), API_OK_VAL

      return "Data type not found.", API_FAILED_VAL

def setupApi(locName):
   app = flask.Flask(locName)
   api = flask_restful.Api(app)

   api.add_resource(DataInfo, "/user/<string:data>")

   return app



'''
Provides a convenient representation of Jupyter's Slider widget and a manager to guarantee that all sliders always add to 1.0
(for weight sliders).
'''
import ipywidgets as widgets

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

MAX_VAL = 1.0
STEP = 0.01

'''
Wraps Jupyter's 'Slider' widget (stored as obj) with a slider description and a boolean keeping track of whether the slider was edited manually or programatically.
'''
class Slider:
   def __init__(self, description, obj):
      self.obj = obj
      self.editedManually = False
      self.description = description
      
   def getReading(self):
      return self.obj.value
   
   def setReading(self, newReading):
      self.obj.value = newReading
      
   '''
   Initializes the slider value given the initial value. This ensures that all
   sliders will start out with equal values.
   '''
   def initSliderValue(self, initVal):
      self.setReading(initVal)

# Manages any number of weight sliders, ensuring that they always add to 1.0 regardless of modification.
class SliderManager:
   '''
   Initializes the slider manager. editedFunction is the function to be triggered when a given slider is edited.
   All subsequent arguments are the sliders to be added to the manager.
   '''
   def __init__(self, editedFunction, slidersList):
      self.editedFunction = editedFunction
      self.slidersList = slidersList
      for slider in slidersList:
         slider.initSliderValue(MAX_VAL / len(slidersList))
         slider.obj.observe(self.sliderEdited, 'value')
        
   # Retrieves a given slider by its description.
   def getSlider(self, description):
      return [slider for slider in self.slidersList if slider.description == description][0]

   # Generates a box widget which displays any number of sliders adjacent to their description.
   def generateDisplayBox(self):
      leftBoxElements = [widgets.Label(slider.description) for slider in self.slidersList]
      rightBoxElements = [slider.obj for slider in self.slidersList]
      return widgets.HBox([widgets.VBox(leftBoxElements), widgets.VBox(rightBoxElements)])

   # Triggers when one of the sliders is edited.
   def sliderEdited(self, change):
      sliders = {slider.obj:slider for slider in self.slidersList}
      editedSlider = change.owner
      deltaVal = change.new - change.old
      diffPerSlider = deltaVal / (len(sliders) - 1)

      isEditingAnything = False
      for s, sliderProperties in sliders.items():
         isEditingAnything = isEditingAnything or sliderProperties.editedManually # determine whether any value in the sliders dictionary is True

      # If the user manually edited a different slider, then this function should exit early because it was triggered
      # by the programmatic editing of this slider. Prevents a recursive loop.
      if isEditingAnything:
         return False

      sliders[editedSlider].editedManually = True

      i = 0

      uneditedSliders = dict(sliders)
      uneditedSliders.pop(editedSlider)

      # Sort the sliders by the amount of distance they can move (the one with the greatest distance should be edited first).
      # If the edited slider was decreased, all the other sliders have to move up, so the one closest to 0.0 should come last.
      # If the edited slider was increased, all the other sliders have to move down, so the one closest to 1.0 should come last. 
      if deltaVal < 0:
         sortedSliders = sorted(uneditedSliders.items(), key = lambda kv: -kv[0].value)
      else:
         sortedSliders = sorted(uneditedSliders.items(), key = lambda kv: kv[0].value - MAX_VAL)

      for s, editing in sortedSliders:
         origVal = s.value
         s.value = max(0, s.value - diffPerSlider)

         # Only subtracts off the difference which was calculated in the slider (useful when the difference brings the
         # slider value to a number less than zero).
         deltaVal -= (origVal - s.value)
         if i != (len(sortedSliders) - 1): # avoid unnecessary division by zero on the last iteration (value will not be used again)
               diffPerSlider = deltaVal / (len(sortedSliders) - i - 1)
         i += 1
   
      # Reset that a value was edited manually to ensure other sliders can trigger this function.
      sliders[editedSlider].editedManually = False
      self.editedFunction()



import datetime

import numpy
from finndex.graphing import gauge
from finndex.util import cryptocurrencies, dateutil, mathutil
from pytrends.request import TrendReq

MIN_TRENDS_VAL = 0
MAX_TRENDS_VAL = 100

def getTrendsDataRaw(keyword, startDate, endDate):
   trends = TrendReq(hl='en-US', tz=0) # tz is timezone offset from UTC in minutes
   trend = trends.get_historical_interest([keyword], 
                                             year_start=startDate.year, month_start=startDate.month, 
                                             day_start=startDate.day, hour_start=startDate.hour, 
                                             year_end=endDate.year, month_end=endDate.month, 
                                             day_end=endDate.day, hour_end=endDate.hour, 
                                             cat=0, geo='', gprop='', sleep=0)[keyword]

   return trend
'''
From trends.google.com: 

Numbers represent search interest relative to the highest point on the chart for the given region and time. 
A value of 100 is the peak popularity for the term. 
A value of 50 means that the term is half as popular. 
A score of 0 means there was not enough data for this term.
'''
def getTrendsData(keyword, startDate, endDate):
   trend = getTrendsDataRaw(keyword, startDate, endDate)

   newDict = {}
   for date, val in trend.items():
      newDict[date] = mathutil.map(val, MIN_TRENDS_VAL, MAX_TRENDS_VAL, 0, 1)

   return newDict

'''
Determines the average trending data for a given keyword on each day within a given range of dates.
Because the Trends API posts multiple values per day, averages each of these results to find a single result
per day.

Returns a dictionary with key as date and value the trends data on that date.
'''
def getTrendsDateRange(startDate, endDate, currenciesList=[cryptocurrencies.Cryptocurrencies.BITCOIN]):
   currenciesDict = {}
   for currency in currenciesList:
      trendsData = getTrendsData(currency.value, startDate, endDate)

      dateDict = {}
      for date, value in trendsData.items():
         if not date in dateDict:
            dateDict[date] = [value]
         else:
            dateDict[date] += [value]
            
      currenciesDict[currency] = {date.date():numpy.average(vals) for date, vals in dateDict.items()}

   return currenciesDict

# Determines the Google trends data on a given date.
def getTrendsDate(date=dateutil.getCurrentDateTime(), keyword="Bitcoin"):
   startDate = datetime.datetime(year=date.year, month=date.month, day=date.day)
   trendsData = getTrendsDataRaw(keyword, startDate, startDate + datetime.timedelta(days=1))
   
   return numpy.average(trendsData)

def displayTrendsDate(date=dateutil.getCurrentDateTime(), display=True, keyword="Bitcoin"):
   return gauge.displayNeutralGauge(getTrendsDate(date=date, keyword=keyword), MIN_TRENDS_VAL, MAX_TRENDS_VAL, "Google Trends", display=display)



from enum import Enum


# Represents an enum of several cryptocurrencies corresponding to their ticker symbols.
class Cryptocurrencies(Enum):
    BITCOIN = "BTC"
    ETHEREUM = "ETH"
    BITCOIN_CASH = "BCH"
    RIPPLE = "XRP"
    LITECOIN = "LTC"
    DOGECOIN = "DOGE"



'''
Provides several utility functions for extracting data from websites (including HTML pages as well as JSON files from APIs).
'''
from contextlib import closing

from bs4 import BeautifulSoup
from requests import get
from requests.exceptions import RequestException

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"


# Extracts the content located at any URL.
def getPageContent(url):
    try: 
        with closing(get(url, stream = True)) as page:
            return page.content.decode("utf-8")
    except RequestException as e:
        print(e)
        return
    
# Parses a string representing HTML, returning the parsed result for convenient iteration.
def parseHTML(url):
    return BeautifulSoup(getPageContent(url), 'html.parser')

# Extracts the text from a CNN article with given URL, excluding the headline and any advertisements.
def getCNNText(url):
    htmlParser = parseHTML(url)
    
    text = ''

    for element in htmlParser.select('div'):
        if element.has_attr('class') and 'zn-body__paragraph' in element['class']: 
            text += element.text 
    text = text.replace('"', ' ')
    return text



from setuptools import setup

setup(name='finndex',
      version='0.2.0',
      description='A useful tool for crypto analysts providing several metrics on various cryptocurrencies.',
      url='https://github.com/FinnitoProductions/Crypto-Sentiment-Tracker',
      download_url='https://github.com/FinnitoProductions/finndex/archive/v0.1.3-alpha.tar.gz',
      author='Finn Frankis',
      author_email='finn@teachmy.com',
      license='MIT',
      packages=['finndex', 'finndex.graphing', 'finndex.util', 'finndex.sentiment', 'finndex.fundamental', 'finndex.aggregate', 'finndex.api'],
      install_requires=['beautifulsoup4',
                        'ipykernel',
                        'ipython',
                        'ipywidgets',
                        'matplotlib',
                        'numpy',
                        'pandas',
                        'pycorenlp',
                        'pytrends',
                        'requests',
                        'stanfordnlp'],
      zip_safe=False)



'''
Manages a set of historical data to plot the aggregate sentiment (with variable weight) relative to other values.
'''

import datetime
import functools
from enum import Enum

import ipywidgets as widgets
import numpy as np
from finndex.aggregate import sliders
from finndex.fundamental import coinmetrics
from finndex.graphing import timeseries
from finndex.sentiment import fearandgreed, trends
from finndex.util import cryptocurrencies, dateutil, mathutil
from IPython.display import display

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

'''
Represents a piece of data with several values over time. 'values' is a dictionary with the date as the key
and the corresponding value on that date as the value. 'slider' is the slider modifying the weight of this reading.
'''
class HistoricalDataReading:
    def __init__(self, name, values, slider):
        self.name = name
        self.values = values
        self.slider = slider

'''
Represents all possible values that can be plotted historically. Each value corresponds to a standard data retrieval function.
'''
class HistoricalMetricType(Enum):
   FEAR_AND_GREED = functools.partial(fearandgreed.getFearAndGreedDateRange)
   TRENDS = functools.partial(trends.getTrendsDateRange)
   BLOCK_COUNT = functools.partial(coinmetrics.getCoinMetricsDateRange, coinmetrics.CoinMetricsData.BLOCK_COUNT)
   TRANSACTION_CNT = functools.partial(coinmetrics.getCoinMetricsDateRange, coinmetrics.CoinMetricsData.TRANSACTION_CNT)
   DAILY_ADDRESSES = functools.partial(coinmetrics.getCoinMetricsDateRange, coinmetrics.CoinMetricsData.DAILY_ADDRESSES)
   MARKET_CAP = functools.partial(coinmetrics.getCoinMetricsDateRange, coinmetrics.CoinMetricsData.MARKET_CAP)
   PRICE_USD = functools.partial(coinmetrics.getCoinMetricsDateRange, coinmetrics.CoinMetricsData.PRICE_USD)

'''
Computes and plots a set of daily historical sentiment values given a set of keywords. Weights can be modified using sliders;
if weights are provided in the 'weights' parameter, presents a static graph using those weights.
'''
class HistoricalSentimentManager:
   def __init__(self, weightedKeywordsList, currenciesList, startDate = dateutil.getCurrentDateTime() - datetime.timedelta(weeks=4), endDate = dateutil.getCurrentDateTime(), weights = None, 
                unweightedKeywordsList = []):
      self.dataDict = {}

      self.startDate = startDate
      self.endDate = endDate
      self.weights=weights
      self.unweightedKeywordsList = unweightedKeywordsList
      self.weightedKeywordsList = weightedKeywordsList
      self.currenciesList = currenciesList

      slidersList = self.computeDataVals()

      if weights == None:
         self.sliderManager = sliders.SliderManager(self.displayGraph, slidersList)
      else:
         self.sliderManager = None

      self.graph = None

   '''
   Generates a list of sliders for and computes the historical values of all desired keywords.
   '''
   def computeDataVals(self):
      slidersList = []
      for idx, keyword in enumerate(self.weightedKeywordsList):
         slider = sliders.Slider(str(keyword), widgets.FloatSlider(min=0.0, max=sliders.MAX_VAL, 
                                                                                             step=sliders.STEP, 
                                                                                             value=self.weights[idx] if self.weights != None else 0.0))
         valuesDict = keyword.value(startDate=self.startDate, endDate=self.endDate, currenciesList=self.currenciesList)

         for currency in self.currenciesList:
            values = valuesDict[currency]

            values = {(date.date() if isinstance(date, datetime.datetime) else date):val for date, val in values.items()}

            dataReading = HistoricalDataReading(name=str(keyword), values=values, 
                                    slider=slider)
            if not currency in self.dataDict:
               self.dataDict[currency] = [dataReading]
            else:
               self.dataDict[currency] += [dataReading]
         
         slidersList += [slider]

      return slidersList

   '''
   Removes any missing data from a weighted dataset by distributing the weight of any missing values (represented by negative infinity)
   to all subsequent values. Does not modify the passed-in list; returns a modified shallow copy.
   '''
   @staticmethod
   def removeGaps(values):
      values = values.copy()
      values.sort() # place all values with -infinity first so they can be dealt with

      # Account for any gaps in the data
      for i, (value, weight) in enumerate(values):
         if value == -float('inf'): # no value provided
            # As long as the last value isn't -infinity (this would mean no data is provided), distribute the
            # weight of the missing data set to all subsequent values
            if i != len(values) - 1:
               weightPerElement = weight / (len(values) - i - 1)
            values[i] = (0.0, 1.0)

            for j, val in enumerate(values[i+1:], start = i+1):
               values[j] = (values[j][0], values[j][1] + weightPerElement)

      return values

   '''
   Computes the weighted historical sentiment with the date as the key and the historical sentiment on that date as the value.
   '''
   def getHistoricalSentiment(self):
      currenciesSentiment = {}
      for currency in self.dataDict:
         sentimentByDate = {}
      
         for date in dateutil.dateRange(self.startDate.date(), self.endDate.date()):
            sentimentByDate[date] = []

            # Populate the sentiment dictionary with date pointing to a list of every corresponding historical sentiment value
            for historicalReading in self.dataDict[currency]:
               sentimentByDate[date] += [(historicalReading.values[date]
                                          if date in historicalReading.values
                                       else -float('inf'), historicalReading.slider.getReading())] # add tuple with value (-infinity if nothing provided) and weight

            values = HistoricalSentimentManager.removeGaps(sentimentByDate[date])

            sentimentByDate[date] = 0

            # Find the weighted average of sentiment on a given day
            for value, weight in values:
               sentimentByDate[date] += value * weight

         currenciesSentiment[currency] = sentimentByDate

      return currenciesSentiment

   '''
   Displays/updates a graph given a dictionary of values 'valueDict' (value type as key and HistoricalDataReading as value).
   If there is no existing graph, generates a new one based on this dictionary. Overlays any additional data specified
   as keyworded-arguments with date as key and the value on that date as the value.

   The values within valueDict are the ones which can be modified in the future. All additional values as keyworded-arguments
   will remain static when plotted and cannot be removed.
   '''
   def displayGraph(self):
      data = {'aggregateSentiment': self.getHistoricalSentiment()}

      # Overlay any additional keywords provided
      for additionalKeyword in self.unweightedKeywordsList:
         value = additionalKeyword.value(startDate = self.startDate, endDate = self.endDate)
         if len(self.unweightedKeywordsList) == 1:
            data[additionalKeyword] = value # If there is only one keyword, the y-axis will be labeled with that keyword
         else:
            ADDITIONAL_DATA_KEY = 'Additional Data' # If there are multiple terms, use a more general term
            if not ADDITIONAL_DATA_KEY in data:
               data[ADDITIONAL_DATA_KEY] = [value]
            else:
               data[ADDITIONAL_DATA_KEY] += [value]
      

      if self.graph == None:
         self.graph = timeseries.TimeSeries(title = "Aggregate Sentiment", data=data, yMin=0, yMax=1)

         if self.sliderManager != None:
            display(self.sliderManager.generateDisplayBox())
      else:
         self.graph.data = data
         self.graph.plotTimeSeries() # updates the graph

      return self.graph
